@@ -12,7 +12,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
 @EXPORT_OK = qw(&render);
 %EXPORT_TAGS = ();
 
-$VERSION = '0.08';
+$VERSION = '0.09';
 
 # Additional valid arguments, fields, and field attributes to those of
 #   HTML::Tabulate
@@ -31,6 +31,8 @@ my %VALID_ARG = (
     # required: list of required/mandatory fields, or tokens 'ALL' or 'NONE'
     required => 'ARRAY/SCALAR',
     # errors: hashref of field => (scalar/array of) validation-error-messages
+    # use_name_as_id: add 'name' as 'id' field to input-type fields if none set
+    use_name_as_id => 'SCALAR',
     errors => 'HASH',
     # errors_where: where to display validation error messages:
     #   top: above form table (default)
@@ -114,6 +116,7 @@ sub init
         hidden => {},
 #       submit => [ 'submit' ],
         xhtml => 1,
+        use_name_as_id => 0,
         null => '&nbsp;',
         errors_where => 'top',
         errors_format => sub {
@@ -279,6 +282,22 @@ sub prerender_munge
 }
 
 
+# -------------------------------------------------------------------------
+# Override start_tag to add explicit 'id' fields if use_name_as_id is set
+#
+sub start_tag
+{
+    my $self = shift;
+    my $tag = shift;
+    my $attr = shift;
+    if ($self->{defn_t}->{use_name_as_id} && 
+        $tag =~ qr/^(input|select|textarea)$/ && 
+        exists $attr->{name}) {
+      $attr->{id} ||= $attr->{name};
+    }
+    return $self->SUPER::start_tag($tag, $attr, @_);
+}
+
 # -------------------------------------------------------------------------
 # Render cells as appropriate input type etc.
 #
@@ -309,7 +328,19 @@ sub cell_content
         if defined $self->{defn_t}->{null} && defined $fattr->{value} &&
             $fattr->{value} eq $self->{defn_t}->{null};
     if ($fattr->{type} eq 'static' || $fattr->{type} eq 'display') {
-        $out .= $fattr->{vlabel} ? sprintf $fattr->{vlabel}, $value : $value;
+        if ($fattr->{vlabel}) {
+          if (ref $fattr->{vlabel}) {
+            if (ref $fattr->{vlabel} eq 'CODE') {
+              $out .= $fattr->{vlabel}->($value, $row, $field);
+            }
+          }
+          else {
+            $out .= sprintf $fattr->{vlabel}, $value;
+          }
+        }
+        else {
+          $out .= $value;
+        }
         delete $fattr->{vlabel};
         $out .= $self->start_tag('input', 
             { type => 'hidden', name => $field, value => $value }, 'close') 
@@ -670,10 +701,10 @@ sub top_errors
     # Report any remaining (presumably non-field-specific) errors
     for my $extra (sort keys %errors) {
         if (ref $errors{$extra} eq 'ARRAY') {
-            push @errors, $_ foreach @{$errors{$extra}};
+            push @errors, sprintf($_, $extra) foreach @{$errors{$extra}};
         }
         else {
-            push @errors, $errors{$extra};
+            push @errors, sprintf($errors{$extra}, $extra);
         }
     }
     return '' unless @errors;
@@ -959,14 +990,33 @@ field => value pairs. Hiddens can also be defined within a field
 attribute section by setting the field type to 'hidden'.
 Default: none.
 
+=item use_name_as_id 
+
+Boolean. If true, HTML::Formulate will add an id attribute set
+to the field name on any input/select/textarea fields that do
+not have an id.
+
 =item errors
 
 Hashref defining a set of field => error_message pairs to be displayed
 as errors on the form (multiple error messages per field are also 
 supported by making the value an arrayref of error messages).
 
-In a similar way to 'required' fields, error field labels are by
-default rendered as:
+Errors are displayed in two ways: the list error messages are 
+error messages is displayed either above the form or in a third 
+column within the form (see 'errors_where' to control which); and
+error field labels are modified to indicate an error.
+
+Error messages are listed in form field order if the error key
+is recognised as a field name ('field errors'); any others are not 
+recognised as field names ('extra errors') are listed after this.
+Error messages are treated as sprintf messages, with a '%s' in the
+message replaced by the field label (for field errors) or the error
+key (for extra errors). Errors without %s placeholders therefore 
+just get rendered as literals.
+
+Field error labels are by default rendered in a similar way to 
+'required' fields, like this:
 
   <th style="color:red"><span class="error_field">Label</span></th>
 
@@ -1098,9 +1148,10 @@ is rendered as the following line:
   </td></tr>
 
 (some formatting newlines added). If you want to use a different label
-than the underlying data value, you can set a (scalar) vlabel, similar
-to selects. The vlabel is interpreted as a sprintf pattern passed the
-current data value i.e. $label = sprintf($vlabel,$value). For example:
+than the underlying data value, you can set a scalar or coderef 'vlabel', 
+similar to selects. A scalar vlabel is interpreted as a sprintf pattern 
+passed the current data value i.e. $label = sprintf($vlabel,$value). 
+For example:
 
   emp_id => { type => 'static', value => '123', vlabel => 'E%05d' }
 
@@ -1112,6 +1163,15 @@ is rendered as:
 
 (newlines added).
 
+A coderef vlabel is passed the standard arguments: value, row, field e.g.
+
+  emp_id => { type => 'static', value => '123', vlabel => sub {
+    my ($value, $row, $field) = @_;
+    sprintf 'E%05d', $value;
+  }}
+
+renders the same as the previous example.
+
 =item omit
 
 The field is to be omitted altogether i.e. no row or input field is to be 
